{
 * This Source Code Form is subject to the terms of the Mozilla Public License,
 * v. 2.0. If a copy of the MPL was not distributed with this file, You can
 * obtain one at http://mozilla.org/MPL/2.0/
 *
 * Copyright (C) 2007-2014, Peter Johnson (www.delphidabbler.com).
 *
 * $Rev: 1957 $
 * $Date: 2014-08-10 15:48:39 +0000 (Sun, 10 Aug 2014) $
 *
 * Classes that encapsulate and executes a command line application and
 * optionally redirects the application's standard input, output and error.
}


unit PJConsoleApp;

{$mode Delphi}

interface


uses
  Classes, LCLIntf, LCLType, Graphics, Types,
{$IF defined(windows)}
  Windows,
{$ELSEIF defined(freebsd) or defined(darwin)}
  ctypes, sysctl,
{$ELSEIF defined(linux)}
{$linklib c}
   ctypes,
{$ENDIF}
  SysUtils, DateUtils, syncobjs;

{$IFDEF Linux}
const _SC_NPROCESSORS_ONLN = 83;
function sysconf(i: cint): clong; cdecl; external name 'sysconf';
{$ENDIF}


const
  // Constants for working in milliseconds
  /// One second in milliseconds.
  cOneSecInMS = 1000;
  /// One minute in milliseconds.
  cOneMinInMS = 60 * cOneSecInMS;

  // Default values for some TPJConsoleApp properties
  /// Default time slice allocated to console app.
  cDefTimeSlice = 50;
  /// Console app's maximum execution time.
  cDefMaxExecTime = cOneMinInMS;

  /// Mask that is ORd with application error codes.
  ///
  ///   According to Windows API docs, error codes with bit 29 set are
  ///  reserved for application use.
  ///   Test for an app error code by anding the error code with this
  ///  mask.
  ///
  cAppErrorMask = 1 shl 29;

  // Application errors
  /// Application time-out error code.
  cAppErrorTimeOut = 1 or cAppErrorMask;
  /// Application terminated error code.
  cAppErrorTerminated = 2 or cAppErrorMask;

/// Checks if given error code is an application defined error.
///
///   Application defined errors have bit 29 set.
function IsApplicationError(const ErrCode: cardinal): Boolean;

type
  ///
  ///  Enumeration of possible priorities for a console application.
  ///
  TPJConsoleAppPriority = (
    cpDefault,    // use default priority (see Win API docs for details)
    cpHigh,       // use for time-critical tasks: processor intensive
    cpNormal,     // normal process with no specific scheduling needs
    cpIdle,       // run only when system idle
    cpRealTime    // highest possible priority: pre-empts all threads inc. OS
  );

type


  ///
  ///  Enumeration of possible colours to be used for a console's foreground and
  ///  background.
  ///
  ///
  ///   It is important that the assigned ordinal values are retained. They
  ///  relate to various combinations of the FOREGROUND_* constants declared in
  ///  the Windows unit.
  ///   Foreground colours are obtained directly from these ordinal values
  ///  while background colours are obtained by left shifting the values by 4.
  ///
  ///   The value names are similar to the equivalent TColor constants, but
  ///  they have different numeric values.
  ///
  TPJConsoleColor = (
    ccBlack   =  0,
    ccNavy    =  1,
    ccGreen   =  2,
    ccTeal    =  3,
    ccMaroon  =  4,
    ccPurple  =  5,
    ccOlive   =  6,
    ccSilver  =  7,
    ccGray    =  8,
    ccBlue    =  9,
    ccLime    = 10,
    ccAqua    = 11,
    ccRed     = 12,
    ccFuchsia = 13,
    ccYellow  = 14,
    ccWhite   = 15
  );

type
  ///
  ///  Type of the TPJConsoleApp.ConsoleColors property. Records the console's
  ///  foreground and background colours.
  ///
  TPJConsoleColors = record
    /// Console's foreground colour.
    Foreground: TPJConsoleColor;
    /// Console's background colour.
    Background: TPJConsoleColor;
  end;

/// "Constructor" function that creates a TPJConsoleColors record from
///  given foreground and background TPJConsoleColor colours.
///   This function is provided because it is not possible to assign the
///  fields of the TPJConsoleApp.ConsoleColors property individually. Instead
///  assign the return value of this function to the property.
function MakeConsoleColors(const AForeground, ABackground: TPJConsoleColor):TPJConsoleColors; overload;

/// "Constructor" function that creates a TPJConsoleColors record from
///  given foreground and background TColor colours.
///
///   This function is provided as a convenience to enable normal TColor
///  values to be used to instantiate a TPJConsoleColors record without having
///  to manually convert to TPJConsoleColor values.
///   An exception is raised if either TColor value is not one of the 16
///  standard colours.
///
function MakeConsoleColors(const AForeground, ABackground: TColor):TPJConsoleColors; overload;

/// "Constructor" function for TSize records.
///   This function is provided because if is not possible to assign the
///  fields of properties of type TSize individually. Instead assign the return
///  value of this function to such properties.
function MakeSize(const ACX, ACY: LongInt): TSize;

type
  ///
  ///  Base class for classes that execute a command line (console) application.
  ///
  ///
  ///   Many properties are available to customise how the console app is
  ///  executed and how the console appears.
  ///   All properties are declared protected. Descendant classes can make
  ///  required properties public.
  ///
  TPJCustomConsoleApp = class(TObject)
  private
    /// Reference to OnWork event handler.
    fOnWork: TNotifyEvent;
    /// Reference to OnComplete event handler.
    fOnComplete: TNotifyEvent;
    /// Reference to OnStart event handler.
    fOnStart: TNotifyEvent;
    /// Handle of console app's redirected standard input. 0 if not
    ///  redirected.
    fStdIn: THandle;
    /// Handle of console app's redirected standard output. 0 if not
    ///  redirected.
    fStdOut: THandle;
    /// Handle of console app's redirected standard error. 0 if not
    ///  redirected.
    fStdErr: THandle;
    /// Command line used to execute application.
    fCommandLine: string;
    /// Specifies console application's current directory.
    fCurrentDir: string;
    /// Exit code returned by console app.
    fExitCode: cardinal;
    /// Maximum execution time of console app (in ms).
    fMaxExecTime: cardinal;
    /// Description of any error that occurred while trying to execute
    ///  console app.
    fErrorMessage: string;
    /// Code of any error that occurred while trying to execute
    ///  console app.
    fErrorCode: cardinal;
    /// Determines whether console app is to be visible or hidden.
    ///
    fVisible: Boolean;
    /// Time console app can run between OnWork events (in ms).
    ///
    fTimeSlice: cardinal;
    /// Determines whether a timed out process is killed.
    fKillTimedOutProcess: Boolean;
    /// Time remaining before console app times out (in ms).
    fTimeToLive: cardinal;
    /// Time since console app started running (in ms).
    fElapsedTime: cardinal;
    /// Flag set by Terminate method to request that a console app is
    ///  terminated.
    fRequestTerminate: Boolean;
    /// Pointer to process security and inheritance attributes.
    ///
    fProcessAttrs: PSecurityAttributes;
    /// Pointer to thread security and inheritance attributes.
    ///
    fThreadAttrs: PSecurityAttributes;
    /// Indicates whether console app should be started in a new
    ///  console window.
    fUseNewConsole: Boolean;
    /// Title to be displayed in a new console window. If '' default
    ///  window title is used.
    fConsoleTitle: string;
    /// Colours to be used in a console window.
    fConsoleColors: TPJConsoleColors;
    /// Size of console screen buffer in character columns and rows.
    ///
    fScreenBufferSize: TSize;
    /// Size of console window in pixels.
    fWindowSize: TSize;
    /// Position of console window in pixel coordinates relative to
    ///  the screen.
    fWindowPosition: TPoint;
    /// Pointer to environment block to be passed to console app.
    ///
    fEnvironment: Pointer;
    /// Flag that indicates if the environment block pointed to by
    ///  fEnvironment is Unicode (True) or ANSI (False).
    fUnicodeEnvironment: Boolean;
    /// Information about running console app process. All values are
    ///  zero when no process is running.
    fProcessInfo: TProcessInformation;
    /// Priority with which console app is started.
    fPriority: TPJConsoleAppPriority;
    /// Monitors a running process, triggering events at the end of
    ///  each timeslice and when the process completes.
    function MonitorProcess: Boolean;
    /// Sets ExitCode property to value returned from console app on
    ///  completion.
    ///  Output: Boolean. True if exit code retrieved OK and False if not.
    ///
    ///   Sets ErrorCode if exit code is not retrieved.
    function SetExitCode: Boolean;
    /// Setter for MaxExecTime property.
    procedure SetMaxExecTime(const Value: cardinal);
    /// Setter for TimeSlice property.
    procedure SetTimeSlice(const Value: cardinal);
    /// Zeroes the process information structure.
    procedure ZeroProcessInfo;
    /// Gets process handle from process info structure.
    ///  Output: THandle. Required process handle or 0 if no process is
    ///  running.
    function GetProcessHandle: THandle;
    /// Updates or clears stored security attributes.
    ///  <"OldValue">PSecurityAttributes [in/out] Passed a pointer
    ///  to attributes to be updated. On return set to point to a copy of
    ///  NewValue structure unless NewValue is nil when OldValue is set to nil.
    ///
    ///  <"NewValue">PSecurityAttributes [in] Pointer to new
    ///  security attributes. May be nil if attributes storage is to be cleared.
    ///
    procedure UpdateSecurityAttrs(var OldValue: PSecurityAttributes; const NewValue: PSecurityAttributes);
    /// Frees memory of a security attributes structure.
    ///  <"Attrs">PSecurityAttributes [in/out] Points to security
    ///  attributes to be freed. Set to nil on return.
    ///   Attrs may be nil, in which case no actions is taken.
    procedure FreeSecurityAttrs(var Attrs: PSecurityAttributes);
    /// Setter for ProcessAttrs property. Makes a copy of new value if
    ///  non-nil.
    procedure SetProcessAttrs(const Value: PSecurityAttributes);
    /// Setter for ThreadAttrs property. Makes a copy of new value if
    ///  non-nil.
    procedure SetThreadAttrs(const Value: PSecurityAttributes);
    
  protected
    /// Starts a process and gets information about it from OS.
    ///
    ///  <"CmdLine">string [in] Command line to be executed.
    ///
    ///  <"CurrentDir">string [in] Application's current directory.
    ///  Pass '' to use same current directory as parent application.
    ///  <"ProcessInfo">TProcessInformation [out] Passes process
    ///  information back to caller.
    ///  Output: Boolean. True if process was created OK, False if process
    ///  couldn't be started.
    function StartProcess(const CmdLine, CurrentDir: string; out ProcessInfo: TProcessInformation): Boolean;
    /// Triggers OnStart event.
    procedure DoStart; virtual;
    /// Triggers OnWork event.
    procedure DoWork; virtual;
    /// Triggers OnComplete event.
    procedure DoComplete; virtual;
    /// Sets error code and message to a class-defined error.
    ///
    ///  <"Code">cardinal [in] Error code. Must have bit 29 set to
    ///  indicate an application error code.
    ///  <"Msg">string [in] Error message.
    procedure RecordAppError(const Code: cardinal; const Msg: string);
    /// Sets error code and message to the last-reported Windows
    ///  error.
    procedure RecordWin32Error;
    /// Resets error code and message to indicate no error.
    procedure ResetError;
    /// Inheritable handle of console app's redirected standard input
    ///  Must be zero if standard input not redirected.
    property StdIn: THandle read fStdIn write fStdIn default 0;
    /// Inheritable handle of console app's redirected standard output
    ///  Must be zero if standard output not redirected.
    property StdOut: THandle read fStdOut write fStdOut default 0;
    /// Inheritable handle of console app's redirected standard error
    ///  Must be zero if standard error not redirected.
    property StdErr: THandle read fStdErr write fStdErr default 0;
    /// Command line to execute. Includes program name and any
    ///  parameters. Paths containing spaces must be quoted.
    property CommandLine: string read fCommandLine write fCommandLine;
    /// Application's current directory. Set to '' to use same current
    ///  directory as parent application.
    property CurrentDir: string read fCurrentDir write fCurrentDir;
    /// Determines whether console app's console is to be displayed
    ///  (True) or hidden (False).
    property Visible: Boolean read fVisible write fVisible default False;
    /// Maximum execution time of console app in ms.
    ///   Set to INFINITE if no execution time limit is required.
    ///
    property MaxExecTime: cardinal read fMaxExecTime write SetMaxExecTime default cDefMaxExecTime;
    /// Time console app executes between OnWork events, in ms.
    ///
    ///
    ///   The app is paused at the end of each time slice while OnWork
    ///  executes.
    ///   Setting TimeSlice to INFINITE means that the app never pauses and
    ///  the OnWork event is never triggered.
    ///
    property TimeSlice: cardinal read fTimeSlice write SetTimeSlice default cDefTimeSlice;
    /// Determines whether timed out console apps are forcible
    ///  terminated.
    ///   When False the Execute method returns when the console app
    ///  times out but leaves the app to run to completion. When true the
    ///  console app is killed when Execute returns.
    property KillTimedOutProcess: Boolean read fKillTimedOutProcess write fKillTimedOutProcess  default True;
    /// Points to a record of security and inheritance attributes for
    ///  a console app process.
    ///
    ///   When nil the process handle can't be inherited.
    ///   When set to a non nil pointer a copy of the referenced structure
    ///  is made.
    ///
    property ProcessAttrs: PSecurityAttributes read fProcessAttrs write SetProcessAttrs default nil;
    /// Points to a record of security and inheritance attributes for
    ///  the console app's primary thread.
    ///
    ///   When nil the thread handle can't be inherited.
    ///   When set to a non nil pointer a copy of the referenced structure
    ///  is made.
    ///
    property ThreadAttrs: PSecurityAttributes read fThreadAttrs write SetThreadAttrs default nil;
    /// Determines if a console app opens a new console window (True)
    ///  or uses any existing console (False).
    property UseNewConsole: Boolean read fUseNewConsole write fUseNewConsole default False;
    /// Title to be displayed in any new console window. If set to the
    ///  empty string the window's default title is displayed.
    ///   If a console app shares a console this property has no effect.
    ///
    property ConsoleTitle: string read fConsoleTitle write fConsoleTitle;
    /// Specifies the foreground and background colours of a new
    ///  console window.
    ///   If a console app shares a console this property has no effect.
    ///
    property ConsoleColors: TPJConsoleColors read fConsoleColors write fConsoleColors;
    /// Specifies the size of a console's screen buffer in character
    ///  columns and rows. If either dimension is zero or negative the default
    ///  buffer size is used.
    ///   If a console app shares a console this property has no effect.
    ///
    property ScreenBufferSize: TSize read fScreenBufferSize write fScreenBufferSize;
    /// Position of top left of console window in pixel co-ordinates
    ///  relative to the screen. If either co-ordinate is negative or zero the
    ///  default window position is used.
    ///   If a console app shares a console this property has no effect.
    ///
    property WindowPosition: TPoint read fWindowPosition write fWindowPosition;
    /// Size of console window in pixels. If either dimension is
    ///  negative and the default window size is used.
    ///   If a console app shares a console this property has no effect.
    ///
    property WindowSize: TSize read fWindowSize write fWindowSize;
    /// Pointer to the environment block to be used by a console app.
    ///
    ///
    ///   If the UnicodeEnvironment property is False then the environment
    ///  block is interpreted as ANSI text but when UnicodeEnvironment is True
    ///  the block is expected to be in Unicode.
    ///   The caller is responsible for allocating and freeing the memory
    ///  used for the environment block. This memory must remain allocated while
    ///  the console app is running.
    ///
    property Environment: Pointer read fEnvironment write fEnvironment;
    /// Flag that indicates if the environment block pointed to by the
    ///  Environment property is in Unicode (True) or in ANSI (False).
    property UnicodeEnvironment: Boolean read fUnicodeEnvironment write fUnicodeEnvironment default False;
    /// Priority with which a console app is executed.
    property Priority: TPJConsoleAppPriority read fPriority write fPriority default cpDefault;
    /// Amount of time, in ms, a console app has remaining before
    ///  it times out.
    ///   This value will be INFINITE if MaxExecTime is INFINITE.
    ///
    property TimeToLive: cardinal read fTimeToLive;
    /// Amount of time, in ms, since a console app started.
    ///   Not updated after the app completes or times out.
    property ElapsedTime: cardinal read fElapsedTime;
    /// Provides information about the executing process.
    ///   All fields are zero when no process is executing.
    property ProcessInfo: TProcessInformation read fProcessInfo;
    /// Records the console app's exit code.
    ///
    ///   The meaning of exit codes is application dependant.
    ///   ExitCode is not valid if ErrorCode is non-zero.
    ///
    property ExitCode: cardinal read fExitCode;
    /// Code indicating if a console app was executed successfully.
    ///  Zero indicates success, non-zero indicates a problem.
    ///   Error codes are either Windows error codes or are set by the
    ///  class (e.g. application timed out or was terminated). Class specific
    ///  error codes have bit 29 set.
    property ErrorCode: cardinal read fErrorCode;
    /// Error message corresponding to the value of ErrorCode. Empty
    ///  string if ErrorCode is zero.
    property ErrorMessage: string read fErrorMessage;
    /// Event triggered just after console application process is
    ///  created, just before it starts executing.
    ///   ProcessInfo has valid data during this event.
    property OnStart: TNotifyEvent read fOnStart write fOnStart;
    /// Event triggered periodically while a console app is executing.
    ///
    ///
    ///   The frequency this event is triggered depends on the value of the
    ///  TimeSlice property. If TimeSlice is INFINITE then the event is never
    ///  triggered.
    ///   ProcessInfo has valid data during this event.
    ///
    property OnWork: TNotifyEvent read fOnWork write fOnWork;
    /// Event triggered when application completes or times out. Can
    ///  be used to tidy up after a process has completed.
    ///
    ///   This event is always triggered. It fires after the last OnWork
    ///  event.
    ///   ErrorCode will have been set and can be used to check how the
    ///  process terminated.
    ///   ProcessInfo has valid data during this event.
    ///
    property OnComplete: TNotifyEvent read fOnComplete write fOnComplete; 
    
  public
    constructor Create;
    destructor Destroy; override;
    
    /// Executes a console app.
    ///  <"CmdLine">string [in] Command line to execute. Includes
    ///  program name and any parameters. Paths containing spaces must be
    ///  quoted.
    ///  <"CurrentDir">string [in] Application's current directory.
    ///  Pass '' to use same current directory as parent application.
    ///  Output: Boolean. True if console app runs successfully or False if
    ///  it fails to run.
    ///   The CommandLine and CurrentDir properties are set to the
    ///  values of the CmdLine and CurrentDir parameters respectively.
    function Execute(const CmdLine: string; const aCurrentDir: string = ''):Boolean; overload;
    /// Executes a console app specified by the CommandLine property
    ///  with the current directory specified by the CurrentDir property.
    ///
    ///  Output: Boolean. True if console app runs successfully or False if
    ///  it fails to run.
    function Execute: Boolean; overload;
    /// Attempts to terminate the current console app.
    ///   Calling this method causes the Execute method to return after
    ///  the next OnWork event. If KillTimedOutProcess is true the console
    ///  application will be halted. The method has no effect when TimeSlice is
    ///  INFINITE.
    procedure Terminate;
  end;


  ///  Class that execute a command line (console) application.
  ///   Many properties are available to customise how the console app is
  ///  executed and how the console appears.
  ///   This class simply makes public all the protected properties of
  ///  TPJCustomConsoleApp.
  ///

  TPJConsoleApp = class(TPJCustomConsoleApp)
  public
    property StdIn;
    property StdOut;
    property StdErr;
    property CommandLine;
    property CurrentDir;
    property Visible;
    property MaxExecTime;
    property TimeSlice;
    property KillTimedOutProcess;
    property ProcessAttrs;
    property ThreadAttrs;
    property UseNewConsole;
    property ConsoleTitle;
    property ConsoleColors;
    property ScreenBufferSize;
    property WindowPosition;
    property WindowSize;
    property Environment;
    property UnicodeEnvironment;
    property Priority;
    property TimeToLive;
    property ElapsedTime;
    property ProcessInfo;
    property ExitCode;
    property ErrorCode;
    property ErrorMessage;
    property OnStart;
    property OnWork;
    property OnComplete;
  end;


implementation


resourcestring
  // Error messages
  sErrTimeout = 'Application timed out';
  sTerminated = 'Application terminated';

function IsApplicationError(const ErrCode: cardinal): Boolean;
begin
  Result := (ErrCode and cAppErrorMask) <> 0;
end;

function MakeConsoleColors(const AForeground, ABackground: TPJConsoleColor):
  TPJConsoleColors;
begin
  Result.Foreground := AForeground;
  Result.Background := ABackground;
end;

function MakeConsoleColors(const AForeground, ABackground: TColor):
  TPJConsoleColors;
  // ---------------------------------------------------------------------------
  // Converts a TColor into equivalent TPJConsoleColor. Raises exception if
  // TColor value is not one of the 16 standard colours.
  function ConvertColor(Color: TColor): TPJConsoleColor;
  const
    sUnsupportedColour = 'Invalid console window colour';
  begin
    case Color of
      // Not using RTL Namespaces means we are using Graphics unit and can use
      // normal clXXX TColor constants
      clBlack:          Result := ccBlack;
      clNavy:           Result := ccNavy;
      clGreen:          Result := ccGreen;
      clTeal:           Result := ccTeal;
      clMaroon:         Result := ccMaroon;
      clPurple:         Result := ccPurple;
      clOlive:          Result := ccOlive;
      clSilver:         Result := ccSilver;
      clGray:           Result := ccGray;
      clBlue:           Result := ccBlue;
      clLime:           Result := ccLime;
      clAqua:           Result := ccAqua;
      clRed:            Result := ccRed;
      clFuchsia:        Result := ccFuchsia;
      clYellow:         Result := ccYellow;
      clWhite:          Result := ccWhite;
      else raise Exception.Create(sUnsupportedColour);
    end;
  end;
  // ---------------------------------------------------------------------------
begin
  Result := MakeConsoleColors(
    ConvertColor(AForeground), ConvertColor(ABackground)
  );
end;

function MakeSize(const ACX, ACY: LongInt): TSize;
begin
  Result.cx := ACX;
  Result.cy := ACY;
end;

{ TPJCustomConsoleApp }

constructor TPJCustomConsoleApp.Create;
begin
  inherited Create;
  // Set default property values
  fMaxExecTime := cDefMaxExecTime;
  fTimeSlice := cDefTimeSlice;
  fVisible := False;
  fStdIn := 0;
  fStdOut := 0;
  fStdErr := 0;
  fKillTimedOutProcess := True;
  ZeroProcessInfo;
  fProcessAttrs := nil;
  fThreadAttrs := nil;
  fUseNewConsole := False;
  fEnvironment := nil;
  fUnicodeEnvironment := False;
  fPriority := cpDefault;
  fConsoleTitle := '';
  fScreenBufferSize := MakeSize(0, 0);
  fConsoleColors := MakeConsoleColors(ccWhite, ccBlack);
  fWindowPosition := types.Point(-1, -1);
  fWindowSize := MakeSize(0, 0);
end;

destructor TPJCustomConsoleApp.Destroy;
begin
  FreeSecurityAttrs(fProcessAttrs);
  FreeSecurityAttrs(fThreadAttrs);
  inherited;
end;

procedure TPJCustomConsoleApp.DoComplete;
begin
  if Assigned(fOnComplete) then
    fOnComplete(Self);
end;

procedure TPJCustomConsoleApp.DoStart;
begin
  if Assigned(fOnStart) then
    fOnStart(Self);
end;

procedure TPJCustomConsoleApp.DoWork;
begin
  if Assigned(fOnWork) then
    fOnWork(Self);
end;

function TPJCustomConsoleApp.Execute: Boolean;
var
  xProcessInfo: TProcessInformation; // information about process
begin
  fExitCode := 0;
  ResetError;
  ZeroProcessInfo;
  Result := StartProcess(fCommandLine, fCurrentDir, xProcessInfo);
  if Result then
  begin
    // Process started: monitor its progress
    try
      fProcessInfo := xProcessInfo;
      DoStart;
      Result := MonitorProcess and SetExitCode;
    finally
      // Process ended: tidy up
      ZeroProcessInfo;
      FileClose(xProcessInfo.hProcess);
      FileClose(xProcessInfo.hThread);
    end;
  end
  else
  begin
    // Couldn't start process: error
    RecordWin32Error;
    ZeroProcessInfo;
  end;
end;

function TPJCustomConsoleApp.Execute(const CmdLine, aCurrentDir: string):
  Boolean;
begin
  fCommandLine := CmdLine;
  fCurrentDir := aCurrentDir;
  Result := Execute;
end;

procedure TPJCustomConsoleApp.FreeSecurityAttrs(var Attrs: PSecurityAttributes);
begin
  if Assigned(Attrs) then
  begin
    FreeMem(Attrs);
    Attrs := nil;
  end;
end;

function TPJCustomConsoleApp.GetProcessHandle: THandle;
begin
  Result := fProcessInfo.hProcess;
end;

function TPJCustomConsoleApp.MonitorProcess: Boolean;
var
  AppState: DWORD;        // State of app after last wait
  StartTime: TDateTime;   // Time application starts
begin
  Result := True;
  StartTime := Now;
  fRequestTerminate := False;
  fTimeToLive := fMaxExecTime;
  fElapsedTime := 0;
  repeat
    // Pause and wait for app - length determined by TimeSlice property
    AppState := WaitForSingleObject(GetProcessHandle, fTimeSlice);
    fElapsedTime := Int64Rec(MilliSecondsBetween(StartTime, Now)).Lo;
    if fMaxExecTime <> INFINITE then
      if fElapsedTime >= fMaxExecTime then
        fTimeToLive := 0
      else
        fTimeToLive := fMaxExecTime - fElapsedTime;
    if AppState = WAIT_FAILED then
    begin
      RecordWin32Error;
      Result := False;
    end
    else if fTimeSlice <> INFINITE then
      // All OK: do inter-timeslice processing
      DoWork;
  until (AppState <> WAIT_TIMEOUT) or (fTimeToLive <= 0) or fRequestTerminate;
  fTimeToLive := 0;
  // App halted or timed out: check which
  if (AppState = WAIT_TIMEOUT) or fRequestTerminate then
  begin
    if fRequestTerminate then
      RecordAppError(cAppErrorTerminated, sTerminated)
    else
      RecordAppError(cAppErrorTimeOut, sErrTimeout);
    DoComplete;   // trigger OnComplete before possibly terminating process
    Result := False;
    if KillTimedOutProcess then
      TerminateProcess(GetProcessHandle, fErrorCode);
  end
  else
    DoComplete;
end;

procedure TPJCustomConsoleApp.RecordAppError(const Code: cardinal;
  const Msg: string);
begin
  Assert(IsApplicationError(Code));
  fErrorCode := Code;
  fErrorMessage := Msg;
end;

procedure TPJCustomConsoleApp.RecordWin32Error;
begin
  fErrorCode := GetLastError;
  fErrorMessage := SysErrorMessage(fErrorCode);
end;

procedure TPJCustomConsoleApp.ResetError;
begin
  fErrorCode := 0;
  fErrorMessage := '';
end;

function TPJCustomConsoleApp.SetExitCode: Boolean;
begin
  Result := GetExitCodeProcess(GetProcessHandle, fExitCode);
  if not Result then
    RecordWin32Error;
end;

procedure TPJCustomConsoleApp.SetMaxExecTime(const Value: cardinal);
begin
  if Value = 0 then
    fMaxExecTime := cDefMaxExecTime
  else
    fMaxExecTime := Value;
end;

procedure TPJCustomConsoleApp.SetProcessAttrs(const Value: PSecurityAttributes);
begin
  UpdateSecurityAttrs(fProcessAttrs, Value);
end;

procedure TPJCustomConsoleApp.SetThreadAttrs(const Value: PSecurityAttributes);
begin
  UpdateSecurityAttrs(fThreadAttrs, Value);
end;

procedure TPJCustomConsoleApp.SetTimeSlice(const Value: cardinal);
begin
  if Value > 0 then
    fTimeSlice := Value
  else
    fTimeSlice := cDefTimeSlice;
end;

function TPJCustomConsoleApp.StartProcess(const CmdLine, CurrentDir: string;
  out ProcessInfo: TProcessInformation): Boolean;
const
  // Maps Visible property to required windows flags
  cShowFlags: array[Boolean] of Integer = (SW_HIDE, SW_SHOW);
  // Maps Priority property to creation flags
  cPriorityFlags: array[TPJConsoleAppPriority] of DWORD = (
    0, HIGH_PRIORITY_CLASS, NORMAL_PRIORITY_CLASS, IDLE_PRIORITY_CLASS,
    REALTIME_PRIORITY_CLASS
  );
var
  StartInfo: TStartupInfo;  // information about process from OS
  CurDir: PChar;            // stores current directory
  CreateFlags: DWORD;       // creation flags
  SafeCmdLine: string;      // stores unique string containing command line
begin
  // Prepare work-around for Unicode CreateProcess API function "feature"
  // See http://bit.ly/adgQ8H
  SafeCmdLine := CmdLine;
  UniqueString(SafeCmdLine);

  // Set up startup information structure
  FillChar(StartInfo, Sizeof(StartInfo),#0);
  with StartInfo do
  begin
    cb := SizeOf(StartInfo);
    dwFlags := STARTF_USESHOWWINDOW or STARTF_USEFILLATTRIBUTE;
    if (fStdIn <> 0) or (fStdOut <> 0) or (fStdErr <> 0) then
      dwFlags := dwFlags or STARTF_USESTDHANDLES;                 // redirecting
    if (fScreenBufferSize.cx > 0) and (fScreenBufferSize.cy > 0) then
    begin
      dwFlags := dwFlags or STARTF_USECOUNTCHARS;  // setting screen buffer size
      dwXCountChars := fScreenBufferSize.cx;
      dwYCountChars := fScreenBufferSize.cy;
    end;
    if (fWindowSize.cx > 0) and (fWindowSize.cy > 0) then
    begin
      dwFlags := dwFlags or STARTF_USESIZE;               // setting window size
      dwXSize := fWindowSize.cx;
      dwYSize := fWindowSize.cy;
    end;
    if (fWindowPosition.X >= 0) and (fWindowPosition.Y >= 0) then
    begin
      dwFlags := dwFlags or STARTF_USEPOSITION;       // setting window position
      dwX := fWindowPosition.X;
      dwY := fWindowPosition.Y;
    end;
    dwFillAttribute := Ord(fConsoleColors.Foreground)   // set fg and bg colours
      or (Ord(fConsoleColors.Background) shl 4);
    if fConsoleTitle <> '' then
      lpTitle := PChar(fConsoleTitle);
    hStdInput := fStdIn;
    hStdOutput := fStdOut;
    hStdError := fStdErr;
    wShowWindow := cShowFlags[fVisible];
  end;

  // Set up process info structure
  ZeroProcessInfo;

  // Set creation flags
  CreateFlags := cPriorityFlags[fPriority];
  if fUseNewConsole then
    CreateFlags := CreateFlags or CREATE_NEW_CONSOLE;
  if fUnicodeEnvironment then
    CreateFlags := CreateFlags or CREATE_UNICODE_ENVIRONMENT;

  // Set current directory
  CurDir := nil;
  if CurrentDir <> '' then
    CurDir := PChar(CurrentDir);

  // Try to create the process
  Result := CreateProcess(
    nil,                  // no application name: we use command line instead
    PChar(SafeCmdLine),   // command line
    fProcessAttrs,        // security attributes for process
    fThreadAttrs,         // security attributes for thread
    True,                 // we inherit inheritable handles from calling process
    CreateFlags,          // creation flags
    fEnvironment,         // environment block for new process
    CurDir,               // current directory
    StartInfo,            // informs how new process' window should appear
    ProcessInfo           // receives info about new process
  );
end;

procedure TPJCustomConsoleApp.Terminate;
begin
  fRequestTerminate := True;
end;

procedure TPJCustomConsoleApp.UpdateSecurityAttrs(
  var OldValue: PSecurityAttributes; const NewValue: PSecurityAttributes);
begin
  if Assigned(NewValue) then
  begin
    if not Assigned(OldValue) then
      GetMem(OldValue, SizeOf(TSecurityAttributes));
    Pointer(OldValue) := Pointer(NewValue);
  end
  else
    FreeSecurityAttrs(OldValue);
end;

procedure TPJCustomConsoleApp.ZeroProcessInfo;
begin
  FillChar(fProcessInfo, SizeOf(fProcessInfo), 0);
end;

end.

