{
 * This Source Code Form is subject to the terms of the Mozilla Public License,
 * v. 2.0. If a copy of the MPL was not distributed with this file, You can
 * obtain one at http://mozilla.org/MPL/2.0/
 *
 * Copyright (C) 2007-2014, Peter Johnson (www.delphidabbler.com).
 *
 * $Rev: 1542 $
 * $Date: 2014-01-11 14:52:19 +0000 (Sat, 11 Jan 2014) $
 *
 * Class that encapsulates an unnamed pipe and can read and write the pipe.
}

unit PJPipe;

{$mode Delphi}

interface


uses
  SysUtils, Classes, LCLIntf, LCLType,
{$IF defined(windows)}
  Windows,
{$ELSEIF defined(freebsd) or defined(darwin)}
  ctypes, sysctl,
{$ELSEIF defined(linux)}
{$linklib c}
   ctypes,
{$ENDIF}
  syncobjs;

type

  ///
  ///  Class that encapsulates an unamed pipe and can read and write the pipe.
  ///
  TPJPipe = class(TObject)
  private
    ///  Handle used to read the pipe.
    fReadHandle: THandle;
    ///  Handle used to write the pipe.
    fWriteHandle: THandle;
    /// Checks that the pipe's write handle hasn't been closed.
    ///
    ///   Raises EInOurError if handle is 0.
    procedure CheckWriteHandle;
    /// Creates a pipe using Windows API and records pipe handles.
    ///
    ///  <"Size">Longword [in] Size of pipe. If 0 the default pipe
    ///  size is used.
    ///  <"Security">PSecurityAttributes [in] Pointer to security
    ///  structure. May be nil.
    ///   Raises EInOutError if pipe can't be created.
    procedure CreatePipe(const Size: LongWord; const Security: PSecurityAttributes);
  public
    /// Object constructor. Creates a pipe with specified size.
    ///
    ///  <"Size">LongWord [in] Required size of pipe. If Size is 0
    ///  then default pipe size used.
    ///  <"Inheritable">Boolean [in] Indicates whether pipe handles
    ///  are to be inheritable.
    ///   Raises EInOutError if pipe can't be created.
    constructor Create(const Size: LongWord; const Inheritable: Boolean = True); overload;
    /// Object constructor. Creates a pipe object with default size.
    ///
    ///  <"Inheritable">Boolean [in] Indicates whether pipe handles
    ///  are to be inheritable.
    ///   Raises EInOutError if pipe can't be created.
    constructor Create(const Inheritable: Boolean = True); overload;
    /// Object constructor. Creates a pipe object with specified size
    ///  and security.
    ///  <"Size">LongWord [in] Required size of pipe. If Size is 0
    ///  then default pipe size used.
    ///  <"Security">TSecurityAttributes [in] Required security for
    ///  pipe. If handles to be inheritable set the bInheritHandle field to
    ///  true.
    ///   Raises EInOutError if pipe can't be created.
    constructor Create(const Size: LongWord; const Security: TSecurityAttributes); overload;
    /// Object constructor. Creates a pipe object with default size
    ///  pipe and specified security.
    ///  <"Security">TSecurityAttributes [in] Required security for
    ///  pipe. If handles to be inheritable set the bInheritHandle field to
    ///  true.
    ///   Raises EInOutError if pipe can't be created.
    constructor Create(const Security: TSecurityAttributes); overload;
    /// Object destructor. Tears down object.
    destructor Destroy; override;
    /// Returns size of data, in bytes, available for reading from
    ///  pipe.
    ///   EInOutError raised if there is an error peeking pipe.
    ///
    function AvailableDataSize: LongWord;
    /// Reads data from pipe into a buffer.
    ///  <"Buf">Untyped [out] Buffer that receives data. Must have
    ///  capacity of at least BufSize bytes.
    ///  <"BufSize">LongWord [in] Size of buffer or number of bytes
    ///  requested.
    ///  <"BytesRead">LongWord [out] Receives number of bytes
    ///  actually read.
    ///  Output: True if some data was read, false if not.
    ///   EInOutError raised if there is an error peeking or reading
    ///  pipe.
    function ReadData(out Buf; const BufSize: LongWord; out BytesRead: LongWord): Boolean;
    /// Reads data from pipe into a byte array.
    ///  <"Count">LongWord [in] Number of bytes to read from pipe. If
    ///  Count = 0 or greater than number of available bytes then all the data
    ///  from the pipe is read.
    ///  Output: Byte array containing data.
    ///   EInOutError raised if there is an error peeking or reading
    ///  pipe.
    function ReadBytes(const Count: LongWord = 0): TBytes;
    /// Copies data from pipe to a stream.
    ///  <"Stm">TStream [in] Stream that receives data.
    ///  <"Count">LongWord [in] Number of bytes to copy. If 0 then
    ///  all remaining data in pipe is copied to stream.
    ///
    ///   EInOutError raised if there is an error peeking or reading pipe.
    ///
    ///   EWriteError raised if fail to write to stream.
    ///
    procedure CopyToStream(const Stm: TStream; Count: LongWord = 0);
    /// Copies data from a stream into the pipe.
    ///  <"Stm">TStream [in] Stream from which to copy data.
    ///  <"Count">LongWord [in] Number of bytes to copy. If 0 then
    ///  all remaining data in stream is copied.
    ///
    ///   EInOutError raised if pipe's write handle has been closed or if
    ///  can't write to pipe.
    ///   EReadError raised if fail to read from stream.
    ///
    procedure CopyFromStream(const Stm: TStream; Count: LongWord = 0);
    /// Writes the whole content of a byte array to pipe.
    ///  <"Bytes">TBytes [in] Array of bytes to write to pipe.
    ///
    ///   EInOutError raised if pipe's write handle has been closed or
    ///  if can't write to pipe.
    procedure WriteBytes(const Bytes: TBytes);
    /// Writes data from buffer to pipe.
    ///  <"Buf">Untyped [in] Buffer containing data to be written.
    ///
    ///  <"BufSize">LongWord [in] Number of bytes to write from
    ///  buffer. Buf must have capacity of at least BufSize bytes.
    ///  Output: Number of bytes actually written.
    ///   EInOutError raised if pipe's write handle has been closed or
    ///  if can't write to pipe.
    function WriteData(const Buf; const BufSize: LongWord): LongWord;
    /// Closes the pipe's write handle if it is open.
    ///   Closing the write handle effectively signals EOF to any reader
    ///  of the pipe. After calling this method no further data may be written
    ///  to the pipe.
    procedure CloseWriteHandle;
    /// Handle used to read data from the pipe. Should be non-zero.
    ///
    property ReadHandle: THandle read fReadHandle;
    /// Handle used to write data to the pipe. Should not be used
    ///  when 0.
    ///   CloseWriteHandle closes and zeros this handle.
    property WriteHandle: THandle read fWriteHandle;
  end;


implementation


Const
  // Error messages
  sCantCreatePipe = 'Can''t create pipe: %s';
  sCantPeekPipe   = 'Can''t read pipe: peek attempt failed';
  sPipeReadError  = 'Error reading pipe';
  sBadWriteHandle = 'Can''t write pipe: handle closed';
  sPipeWriteError = 'Error writing pipe';


{ TPJPipe }

function TPJPipe.AvailableDataSize: LongWord;
begin
  if not PeekNamedPipe(fReadHandle, nil, 0, nil, @Result, nil) then
    raise EInOutError.Create(sCantPeekPipe);
end;

procedure TPJPipe.CheckWriteHandle;
begin
  if fWriteHandle = 0 then
    raise EInOutError.Create(sBadWriteHandle);
end;

procedure TPJPipe.CloseWriteHandle;
begin
  if fWriteHandle <> 0 then
  begin
    FileClose(fWriteHandle); { *Converted from CloseHandle* }
    fWriteHandle := 0;
  end;
end;

procedure TPJPipe.CopyFromStream(const Stm: TStream; Count: LongWord);
var
  BytesToWrite: LongWord;       // adjusted number of bytes to write
  Buf: array[0..4095] of Byte;  // buffer used in copying from pipe to stream
begin
  // Determine how much to copy
  if Count = 0 then
    Count := Stm.Size - Stm.Position;
  // Copy data one bufferful at a time
  while Count > 0 do
  begin
    if Count > SizeOf(Buf) then
      BytesToWrite := SizeOf(Buf)
    else
      BytesToWrite := Count;
    Stm.ReadBuffer(Buf, BytesToWrite);
    WriteData(Buf, BytesToWrite);
    Dec(Count, BytesToWrite);
  end;
end;

procedure TPJPipe.CopyToStream(const Stm: TStream; Count: LongWord);
var
  AvailBytes: LongWord;           // number of bytes in pipe
  BytesToRead: LongWord;          // decreasing count of remaining bytes
  BytesRead: LongWord;            // bytes read in each loop
  Buf: array[0..4095] of Byte;    // buffer used to read from stream
begin
  // Determine how much should be read
  AvailBytes := AvailableDataSize;
  if (Count = 0) or (Count > AvailBytes) then
    Count := AvailBytes;
  // Copy data one bufferful at a time
  while Count > 0 do
  begin
    if Count > SizeOf(Buf) then
      BytesToRead := SizeOf(Buf)
    else
      BytesToRead := Count;
    ReadData(Buf, BytesToRead, BytesRead);
    if BytesRead <> BytesToRead then
      raise EInOutError.Create(sPipeReadError);
    Stm.WriteBuffer(Buf, BytesRead);
    Dec(Count, BytesRead);
  end;
end;

constructor TPJPipe.Create(const Size: LongWord; const Inheritable: Boolean);
var
  Security: TSecurityAttributes;  // file's security attributes
begin
  inherited Create;
  if Inheritable then
  begin
    // Set up security structure so file handle is inheritable (for Windows NT)
    Security.nLength := SizeOf(Security);
    Security.lpSecurityDescriptor := nil;
    Security.bInheritHandle := True;
    CreatePipe(Size, @Security);
  end
  else
    CreatePipe(Size, nil);
end;

constructor TPJPipe.Create(const Size: LongWord;
  const Security: TSecurityAttributes);
begin
  inherited Create;
  CreatePipe(Size, @Security);
end;

constructor TPJPipe.Create(const Security: TSecurityAttributes);
begin
  Create(0, Security);
end;

constructor TPJPipe.Create(const Inheritable: Boolean);
begin
  Create(0, Inheritable);
end;

procedure TPJPipe.CreatePipe(const Size: LongWord;
  const Security: PSecurityAttributes);
begin
  if not Windows.CreatePipe(fReadHandle, fWriteHandle, Security, Size) then
    raise EInOutError.CreateFmt(
      sCantCreatePipe, [SysErrorMessage(GetLastError)]
    );
end;

destructor TPJPipe.Destroy;
begin
  FileClose(fReadHandle); { *Converted from CloseHandle* }
  CloseWriteHandle;
  inherited;
end;

function TPJPipe.ReadBytes(const Count: LongWord): TBytes;
var
  AvailBytes: LongWord; // number of bytes in pipe
  BytesRead: LongWord;  // number of bytes actually read from pipe
begin
  AvailBytes := AvailableDataSize;
  if (Count = 0) or (Count > AvailBytes) then
    SetLength(Result, AvailBytes)
  else
    SetLength(Result, Count);
  ReadData(Pointer(Result)^, Length(Result), BytesRead);
  if BytesRead <> LongWord(Length(Result)) then
    raise EInOutError.Create(sPipeReadError);
end;

function TPJPipe.ReadData(out Buf; const BufSize: LongWord;
  out BytesRead: LongWord): Boolean;
var
  BytesToRead: DWORD;   // number of bytes to actually read
begin
  BytesToRead := AvailableDataSize;
  if BytesToRead > 0 then
  begin
    if BytesToRead > BufSize then
      BytesToRead := BufSize;
    // we don't check return value: sometimes fails with BytesRead = 0
    ReadFile(fReadHandle, Buf, BytesToRead, BytesRead, nil);
    Result := BytesRead > 0;
  end
  else
  begin
    Result := False;
    BytesRead := 0;
  end;
end;

procedure TPJPipe.WriteBytes(const Bytes: TBytes);
begin
  WriteData(Pointer(Bytes)^, Length(Bytes));
end;

function TPJPipe.WriteData(const Buf; const BufSize: LongWord): LongWord;
begin
  CheckWriteHandle;
  if not WriteFile(fWriteHandle, Buf, BufSize, Result, nil) then
    raise EInOutError.Create(sPipeWriteError);
end;

end.

