{
 * This Source Code Form is subject to the terms of the Mozilla Public License,
 * v. 2.0. If a copy of the MPL was not distributed with this file, You can
 * obtain one at http://mozilla.org/MPL/2.0/
 *
 * Copyright (C) 2011-2014, Peter Johnson (www.delphidabbler.com).
 *
 * $Rev: 1515 $
 * $Date: 2014-01-11 02:36:28 +0000 (Sat, 11 Jan 2014) $
 *
 * Classes that filter or format output from pipes.
}

unit PJPipeFilters;

{$mode Delphi}

interface

uses
  SysUtils, Classes,
  PJPipe;

type

  ///  Abstract base class for all classes that filter data read from pipes.

  TPJPipeFilter = class(TObject)
  private
    ///  Pipe that provides data
    fPipe: TPJPipe;
    ///  Indicates if pipe is owned (and freed) by this object
    fOwnsPipe: Boolean;
    ///  Buffer containing left over data not processed after last pipe read.
    fUnprocessedData: TBytes;
  protected
    /// Returns number of bytes at end of RawData that cannot be
    ///  processed yet.
    ///   Any such bytes are stored for use on next pipe read.
    function LeftOverByteCount(const RawData: TBytes): Integer; virtual;
      abstract;
    /// Performs filtering operation on given data.
    procedure DoFilter(const ValidData: TBytes); virtual; abstract;
    /// Flushes any remaining data, assuming there is no more to come.
    ///
    procedure DoFlush; virtual; abstract;
    /// Processes given new Data.
    procedure ProcessData(const Data: TBytes);
  public
    /// Object constructor. Records parameters.
    ///  <"APipe">TPJPipe [in] Pipe that provides data to be
    ///  filtered.
    ///  <"AOwnsPipe">Boolean [in] Flag true if this object is to
    ///  own (and free) APipe.
    constructor Create(const APipe: TPJPipe; const AOwnsPipe: Boolean = False);
    ///  <summry>Object destructor. Tears down object and performs flush
    ///  operation.
    destructor Destroy; override;
    /// Reads all available data from pipe.
    procedure ReadPipe;
    /// Flushes any unprocessed data.
    procedure Flush;
    /// Checks if object currently contains any un processed data.
    ///
    ///   Call this after last call to AddData to test if there is any
    ///  data left un-processed. If so the data stream is not valid Unicode.
    ///
    function HaveUnprocessedData: Boolean;
    /// Read only property providing access to pipe that is being
    ///  filtered.
    property Pipe: TPJPipe read fPipe;
  end;

type
  /// Type of text read event triggered by classes that read Unicode
  ///  strings from pipes.
  ///  <"Sender">TObject [in] Object that triggers event.
  ///  <"Text">UnicodeString [in] Text for which event was triggered.
  ///
  ///   Type of OnText and OnLineEnd events.
  TPJUnicodeTextReadEvent = procedure(Sender: TObject;
    const Text: UnicodeString) of object;

type
  ///
  ///  Constructs valid Unicode BMP strings from chunks of data read from pipe.
  ///  Chunk may or may not be split on valid Unicode character boundaries. An
  ///  event is triggered for each string read. Strings are also parsed into
  ///  lines that are split by a specified end of line marker.
  ///
  ///
  ///   The class only works for Unicode text from the basic multingual
  ///  plane (BMP).
  ///   For info about the BMP see http://bit.ly/f3Ardu
  ///
  TPJUnicodeBMPPipeFilter = class(TPJPipeFilter)
  private
    ///  Text data buffered awaiting next end of line.
    fPartialLine: UnicodeString;
    ///  Stores reference to OnText event handler.
    fOnText: TPJUnicodeTextReadEvent;
    ///  Stores reference to OnLineEnd event handler.
    fOnLineEnd: TPJUnicodeTextReadEvent;
    ///  Value of EOLMarker property.
    fEOLMarker: UnicodeString;
    /// Triggers OnText event for a given Unicode string.
    procedure DoText(const Text: UnicodeString);
    /// Triggers OnLineEnd event for a given Unicode string.
    procedure DoLineEnd(const Text: UnicodeString);
  protected
    /// Returns number of bytes at end of RawData that cannot be
    ///  processed yet.
    ///   Bytes that can't be processed are fragments of a Unicode
    ///  character.
    function LeftOverByteCount(const RawData: TBytes): Integer; override;
    /// Extracts Unicode text from given data and parses it into
    ///  lines.
    procedure DoFilter(const ValidData: TBytes); override;
    /// Flushes any un-reported line of text, triggering OnLineEnd
    ///  event for it.
    procedure DoFlush; override;
  public
    /// Sets default property values.
    procedure AfterConstruction; override;
    /// Character(s) to be used as end of line marker used when
    ///  parsing text into lines.
    ///   Defaults to CRLF (#13#10) or (#10).
    property EOLMarker: UnicodeString read fEOLMarker write fEOLMarker;
    /// Event triggered whenever valid text is read from pipe.
    ///  Contains text up to and including last valid Unicode character in given
    ///  data.
    property OnText: TPJUnicodeTextReadEvent read fOnText write fOnText;
    /// Event triggered when each end of line is reached. Also
    ///  triggered for any pending text when Flush is called, or object is
    ///  destroyed.
    property OnLineEnd: TPJUnicodeTextReadEvent read fOnLineEnd write fOnLineEnd;
  end;

type
  /// Type of text read event triggered by classes that read Ansi
  ///  strings from pipes.
  ///  <"Sender">TObject [in] Object that triggers event.
  ///  <"Text">AnsiString [in] Text for which event was triggered.
  ///
  ///   Type of OnText and OnLineEnd events.
  TPJAnsiTextReadEvent = procedure(Sender: TObject; const Text: AnsiString) of object;

type
  ///
  ///  Constructs valid ANSI SBCS strings from chunks of data read from pipe. An
  ///  event is triggered for each string read. Strings are also parsed into
  ///  lines that are split by a specified end of line marker.
  ///
  ///
  ///   The class only works for Single Byte Character Set (SBCS) strings,
  ///  i.e. those that take exactly one byte per character. It is the caller's
  ///  responsibility to interpret the string's code page correctly.
  ///   For info about SBCS see http://bit.ly/epPKtW
  ///
  TPJAnsiSBCSPipeFilter = class(TPJPipeFilter)
  private
    ///  Text data buffered awaiting next end of line.
    fPartialLine: AnsiString;
    ///  Stores reference to OnText event handler.
    fOnText: TPJAnsiTextReadEvent;
    ///  Stores reference to OnLineEnd event handler.
    fOnLineEnd: TPJAnsiTextReadEvent;
    ///  Value of EOLMarker property.
    fEOLMarker: AnsiString;
    /// Triggers OnText event for a given ANSI string.
    procedure DoText(const Text: AnsiString);
    /// Triggers OnLineEnd event for a given ANSI string.
    procedure DoLineEnd(const Text: AnsiString);
  protected
    /// Returns number of bytes at end of RawData that cannot be
    ///  processed yet.
    ///   All bytes can always be processed in a SBCS string, so this
    ///  method always returns 0.
    function LeftOverByteCount(const RawData: TBytes): Integer; override;
    /// Extracts ANSI text from given data and parses it into lines.
    ///
    procedure DoFilter(const ValidData: TBytes); override;
    /// Flushes any un-reported line of text, triggering OnLineEnd
    ///  event for it.
    procedure DoFlush; override;
  public
    /// Sets default property values.
    procedure AfterConstruction; override;
    /// Character(s) to be used as end of line marker used when
    ///  parsing text into lines.
    ///   Defaults to CRLF (#13#10).
    property EOLMarker: AnsiString read fEOLMarker write fEOLMarker;
    /// Event triggered whenever text is read from pipe. Contains text
    ///  up to and including last ANSI character in given data.
    property OnText: TPJAnsiTextReadEvent read fOnText write fOnText;
    /// Event triggered when each end of line is reached. Also
    ///  triggered for any pending text when Flush is called, or object is
    ///  destroyed.
    property OnLineEnd: TPJAnsiTextReadEvent read fOnLineEnd write fOnLineEnd;
  end;

implementation

Const
{$IFDEF WINDOWS}
  cnEOLMarker = #13#10;
{$ELSE}
  cnEOLMarker = #10;
{$ENDIF}

//==== TPJPipeFilter ===========================

constructor TPJPipeFilter.Create(const APipe: TPJPipe;
  const AOwnsPipe: Boolean);
begin
  Assert(Assigned(APipe));
  inherited Create;
  fPipe := APipe;
  fOwnsPipe := AOwnsPipe;
end;

destructor TPJPipeFilter.Destroy;
begin
  DoFlush;
  if fOwnsPipe then
    fPipe.Free;
  inherited;
end;

procedure TPJPipeFilter.Flush;
begin
  DoFlush;
end;

function TPJPipeFilter.HaveUnprocessedData: Boolean;
begin
  Result := Length(fUnprocessedData) > 0;
end;

procedure TPJPipeFilter.ProcessData(const Data: TBytes);
var
  Buffer: TBytes;
  TrimmedByteCount: Integer;
begin
  if Length(Data) = 0 then
    Exit;
  SetLength(Buffer, Length(fUnprocessedData) + Length(Data));
  if Length(fUnprocessedData) > 0 then
    Move(fUnprocessedData[0], Buffer[0], Length(fUnprocessedData));
  Move(Data[0], Buffer[Length(fUnprocessedData)], Length(Data));
  TrimmedByteCount := LeftOverByteCount(Buffer);
  if TrimmedByteCount > 0 then
  begin
    fUnprocessedData := Copy(
      Buffer, Length(Buffer) - TrimmedByteCount, TrimmedByteCount
    );
    SetLength(Buffer, Length(Buffer) - TrimmedByteCount);
  end
  else
    SetLength(fUnprocessedData, 0);
  DoFilter(Buffer);
end;

procedure TPJPipeFilter.ReadPipe;
begin
  ProcessData(fPipe.ReadBytes(0));
end;

//==== TPJUnicodeBMPPipeFilter =======================

procedure TPJUnicodeBMPPipeFilter.AfterConstruction;
begin
  inherited;
  fEOLMarker := cnEOLMarker;
end;

procedure TPJUnicodeBMPPipeFilter.DoFilter(const ValidData: TBytes);
var
  Text: UnicodeString;  // Text read from ValidData
  EOLPos: Integer;      // Position(s) of end of line marker in text
begin
  // each character is exactly SizeOf(WideChar) bytes in basic multilingual
  // plane
  SetLength(Text, Length(ValidData) div SizeOf(WideChar));
  Move(Pointer(ValidData)^, Pointer(Text)^, Length(ValidData));
  // Trigger OnText event for read text
  DoText(Text);
  fPartialLine := fPartialLine + Text;
  // Break text into lines separated by EOLMarker, trigger OnLineEnd for each
  EOLPos := Pos(fEOLMarker, fPartialLine);
  while EOLPos > 0 do
  begin
    DoLineEnd(Copy(fPartialLine, 1, EOLPos - 1));
    fPartialLine := Copy(
      fPartialLine, EOLPos + Length(fEOLMarker), MaxInt
    );
    EOLPos := Pos(fEOLMarker, fPartialLine);
  end;
end;

procedure TPJUnicodeBMPPipeFilter.DoFlush;
begin
  if fPartialLine <> '' then
  begin
    DoLineEnd(fPartialLine);
    fPartialLine := '';
  end;
end;

procedure TPJUnicodeBMPPipeFilter.DoLineEnd(const Text: UnicodeString);
begin
  if Assigned(OnLineEnd) then
    OnLineEnd(Self, Text);
end;

procedure TPJUnicodeBMPPipeFilter.DoText(const Text: UnicodeString);
begin
  if Assigned(OnText) then
    OnText(Self, Text);
end;

function TPJUnicodeBMPPipeFilter.LeftOverByteCount(
  const RawData: TBytes): Integer;
begin
  Result := Length(RawData) mod SizeOf(WideChar);
end;

{ TPJAnsiSBCSPipeFilter }

procedure TPJAnsiSBCSPipeFilter.AfterConstruction;
begin
  inherited;
  fEOLMarker := cnEOLMarker;
end;

procedure TPJAnsiSBCSPipeFilter.DoFilter(const ValidData: TBytes);
var
  Text: AnsiString;  // Text read from ValidData
  EOLPos: Integer;   // Position(s) of end of line marker in text
begin
  // NOTE: code assumes SizeOf(AnsiChar) = 1
  // Record ANSI SBCS text
  SetLength(Text, Length(ValidData));
  Move(Pointer(ValidData)^, Pointer(Text)^, Length(ValidData));
  // Trigger OnText event for read text
  DoText(Text);
  fPartialLine := fPartialLine + Text;
  // Break text into lines separated by EOLMarker, trigger OnLineEnd for each
  EOLPos := Pos(fEOLMarker, fPartialLine);
  while EOLPos > 0 do
  begin
    DoLineEnd(Copy(fPartialLine, 1, EOLPos - 1));
    fPartialLine := Copy(
      fPartialLine, EOLPos + Length(fEOLMarker), MaxInt
    );
    EOLPos := Pos(fEOLMarker, fPartialLine);
  end;
end;

procedure TPJAnsiSBCSPipeFilter.DoFlush;
begin
  if fPartialLine <> '' then
  begin
    DoLineEnd(fPartialLine);
    fPartialLine := '';
  end;
end;

procedure TPJAnsiSBCSPipeFilter.DoLineEnd(const Text: AnsiString);
begin
  if Assigned(OnLineEnd) then
    OnLineEnd(Self, Text);
end;

procedure TPJAnsiSBCSPipeFilter.DoText(const Text: AnsiString);
begin
  if Assigned(OnText) then
    OnText(Self, Text);
end;

function TPJAnsiSBCSPipeFilter.LeftOverByteCount(
  const RawData: TBytes): Integer;
begin
  // all bytes are valid because each character is exactly one byte in a SBCS
  Result := 0;
end;

end.

